@@ -47,6 +47,8 @@ struct _MutterWindowPrivate
    * texture */
   GdkRegion        *bounding_region;
 
+  gint              freeze_count;
+
   /*
    * These need to be counters rather than flags, since more plugins
    * can implement same effect; the practicality of stacking effects
@@ -65,7 +67,9 @@ struct _MutterWindowPrivate
   guint		    disposed               : 1;
   guint             redecorating           : 1;
 
-  guint		    needs_repair           : 1;
+  guint		    needs_damage_all       : 1;
+
+  guint		    needs_pixmap           : 1;
   guint		    needs_reshape          : 1;
   guint		    size_changed           : 1;
 
@@ -272,9 +276,6 @@ mutter_meta_window_decorated_notify (MetaWindow *mw,
       priv->damage = None;
     }
 
-  g_hash_table_remove (info->windows_by_xid, (gpointer) priv->xwindow);
-  g_hash_table_insert (info->windows_by_xid, (gpointer) new_xwindow, self);
-
   g_free (priv->desc);
   priv->desc = NULL;
 
@@ -335,7 +336,8 @@ mutter_window_constructed (GObject *object)
   if (priv->attrs.class == InputOnly)
     priv->damage = None;
   else
-    priv->damage = XDamageCreate (xdisplay, xwindow, XDamageReportNonEmpty);
+    priv->damage = XDamageCreate (xdisplay, xwindow,
+                                  XDamageReportRawRectangles);
 
   format = XRenderFindVisualFormat (xdisplay, priv->attrs.visual);
 
@@ -421,7 +423,6 @@ mutter_window_dispose (GObject *object)
     }
 
   info->windows = g_list_remove (info->windows, (gconstpointer) self);
-  g_hash_table_remove (info->windows_by_xid, (gpointer) priv->xwindow);
 
   /*
    * Release the extra reference we took on the actor.
@@ -769,6 +770,60 @@ mutter_window_showing_on_its_workspace (MutterWindow *self)
   return meta_window_showing_on_its_workspace (self->priv->window);
 }
 
+static void
+mutter_window_freeze (MutterWindow *self)
+{
+  self->priv->freeze_count++;
+}
+
+static void
+mutter_window_damage_all (MutterWindow *self)
+{
+  MutterWindowPrivate *priv = self->priv;
+  ClutterX11TexturePixmap *texture_x11 = CLUTTER_X11_TEXTURE_PIXMAP (priv->actor);
+  guint pixmap_width = 0;
+  guint pixmap_height = 0;
+
+  if (!priv->needs_damage_all)
+    return;
+
+  g_object_get (texture_x11,
+                "pixmap-width", &pixmap_width,
+                "pixmap-height", &pixmap_height,
+                NULL);
+
+  clutter_x11_texture_pixmap_update_area (texture_x11,
+                                          0,
+                                          0,
+                                          pixmap_width,
+                                          pixmap_height);
+
+  priv->needs_damage_all = FALSE;
+}
+
+static void
+mutter_window_thaw (MutterWindow *self)
+{
+  self->priv->freeze_count--;
+
+  if (G_UNLIKELY (self->priv->freeze_count < 0))
+    {
+      g_warning ("Error in freeze/thaw accounting.");
+      self->priv->freeze_count = 0;
+      return;
+    }
+
+  if (self->priv->freeze_count)
+    return;
+
+  /* Since we ignore damage events while a window is frozen for certain effects
+   * we may need to issue an update_area() covering the whole pixmap if we
+   * don't know what real damage has happened. */
+
+  if (self->priv->needs_damage_all)
+    mutter_window_damage_all (self);
+}
+
 gboolean
 mutter_window_effect_in_progress (MutterWindow *self)
 {
@@ -780,11 +835,11 @@ mutter_window_effect_in_progress (MutterWindow *self)
 }
 
 static void
-mutter_window_mark_for_repair (MutterWindow *self)
+mutter_window_queue_create_pixmap (MutterWindow *self)
 {
   MutterWindowPrivate *priv = self->priv;
 
-  priv->needs_repair = TRUE;
+  priv->needs_pixmap = TRUE;
 
   if (!priv->mapped)
     return;
@@ -800,6 +855,21 @@ mutter_window_mark_for_repair (MutterWindow *self)
   clutter_actor_queue_redraw (priv->actor);
 }
 
+static gboolean
+is_freeze_thaw_effect (gulong event)
+{
+  switch (event)
+  {
+  case MUTTER_PLUGIN_DESTROY:
+  case MUTTER_PLUGIN_MAXIMIZE:
+  case MUTTER_PLUGIN_UNMAXIMIZE:
+    return TRUE;
+    break;
+  default:
+    return FALSE;
+  }
+}
+
 static gboolean
 start_simple_effect (MutterWindow *self,
                      gulong        event)
@@ -807,6 +877,7 @@ start_simple_effect (MutterWindow *self,
   MutterWindowPrivate *priv = self->priv;
   MetaCompScreen *info = meta_screen_get_compositor_data (priv->screen);
   gint *counter = NULL;
+  gboolean use_freeze_thaw = FALSE;
 
   if (!info->plugin_mgr)
     return FALSE;
@@ -831,6 +902,11 @@ start_simple_effect (MutterWindow *self,
 
   g_assert (counter);
 
+  use_freeze_thaw = is_freeze_thaw_effect (event);
+
+  if (use_freeze_thaw)
+    mutter_window_freeze (self);
+
   (*counter)++;
 
   if (!mutter_plugin_manager_event_simple (info->plugin_mgr,
@@ -838,6 +914,8 @@ start_simple_effect (MutterWindow *self,
                                            event))
     {
       (*counter)--;
+      if (use_freeze_thaw)
+        mutter_window_thaw (self);
       return FALSE;
     }
 
@@ -861,7 +939,7 @@ mutter_window_after_effects (MutterWindow *self)
   if (!meta_window_is_mapped (priv->window))
     mutter_window_detach (self);
 
-  if (priv->needs_repair)
+  if (priv->needs_pixmap)
     clutter_actor_queue_redraw (priv->actor);
 }
 
@@ -930,6 +1008,9 @@ mutter_window_effect_completed (MutterWindow *self,
     break;
   }
 
+  if (is_freeze_thaw_effect (event))
+    mutter_window_thaw (self);
+
   if (!mutter_window_effect_in_progress (self))
     mutter_window_after_effects (self);
 }
@@ -955,7 +1036,7 @@ mutter_window_detach (MutterWindow *self)
   clutter_x11_texture_pixmap_set_pixmap (CLUTTER_X11_TEXTURE_PIXMAP (priv->actor),
                                          None);
 
-  mutter_window_mark_for_repair (self);
+  mutter_window_queue_create_pixmap (self);
 }
 
 void
@@ -976,7 +1057,6 @@ mutter_window_destroy (MutterWindow *self)
    */
   info = meta_screen_get_compositor_data (priv->screen);
   info->windows = g_list_remove (info->windows, (gconstpointer) self);
-  g_hash_table_remove (info->windows_by_xid, (gpointer)priv->xwindow);
 
   if (priv->type == META_COMP_WINDOW_DROPDOWN_MENU ||
       priv->type == META_COMP_WINDOW_POPUP_MENU ||
@@ -1019,7 +1099,7 @@ mutter_window_sync_actor_position (MutterWindow *self)
       priv->attrs.height != window_rect.height)
     {
       priv->size_changed = TRUE;
-      mutter_window_mark_for_repair (self);
+      mutter_window_queue_create_pixmap (self);
     }
 
   /* XXX deprecated: please use meta_window_get_outer_rect instead */
@@ -1136,6 +1216,7 @@ mutter_window_maximize (MutterWindow       *self,
   clutter_actor_set_size (CLUTTER_ACTOR (self), old_rect->width, old_rect->height);
 
   self->priv->maximize_in_progress++;
+  mutter_window_freeze (self);
 
   if (!info->plugin_mgr ||
       !mutter_plugin_manager_event_maximize (info->plugin_mgr,
@@ -1146,6 +1227,7 @@ mutter_window_maximize (MutterWindow       *self,
 
     {
       self->priv->maximize_in_progress--;
+      mutter_window_thaw (self);
     }
 }
 
@@ -1163,6 +1245,7 @@ mutter_window_unmaximize (MutterWindow      *self,
   clutter_actor_set_size (CLUTTER_ACTOR (self), old_rect->width, old_rect->height);
 
   self->priv->unmaximize_in_progress++;
+  mutter_window_freeze (self);
 
   if (!info->plugin_mgr ||
       !mutter_plugin_manager_event_maximize (info->plugin_mgr,
@@ -1172,6 +1255,7 @@ mutter_window_unmaximize (MutterWindow      *self,
 					     new_rect->width, new_rect->height))
     {
       self->priv->unmaximize_in_progress--;
+      mutter_window_thaw (self);
     }
 }
 
@@ -1210,6 +1294,8 @@ mutter_window_new (MetaWindow *window)
   priv = self->priv;
 
   priv->mapped = meta_window_toplevel_is_mapped (priv->window);
+  if (priv->mapped)
+    mutter_window_queue_create_pixmap (self);
 
   mutter_window_sync_actor_position (self);
 
@@ -1224,7 +1310,6 @@ mutter_window_new (MetaWindow *window)
    * before we first paint.
    */
   info->windows = g_list_append (info->windows, self);
-  g_hash_table_insert (info->windows_by_xid, (gpointer) top_window, self);
 
   return self;
 }
@@ -1238,7 +1323,7 @@ mutter_window_mapped (MutterWindow *self)
 
   priv->mapped = TRUE;
 
-  mutter_window_mark_for_repair (self);
+  mutter_window_queue_create_pixmap (self);
 }
 
 void
@@ -1254,7 +1339,7 @@ mutter_window_unmapped (MutterWindow *self)
     return;
 
   mutter_window_detach (self);
-  priv->needs_repair = FALSE;
+  priv->needs_pixmap = FALSE;
 }
 
 static void
@@ -1466,7 +1551,7 @@ mutter_window_reset_visible_regions (MutterWindow *self)
 }
 
 static void
-check_needs_repair (MutterWindow *self)
+check_needs_pixmap (MutterWindow *self)
 {
   MutterWindowPrivate *priv     = self->priv;
   MetaScreen          *screen   = priv->screen;
@@ -1477,7 +1562,7 @@ check_needs_repair (MutterWindow *self)
   Window               xwindow  = priv->xwindow;
   gboolean             full     = FALSE;
 
-  if (!priv->needs_repair)
+  if (!priv->needs_pixmap)
     return;
 
   if (!priv->mapped)
@@ -1549,74 +1634,50 @@ check_needs_repair (MutterWindow *self)
       full = TRUE;
     }
 
- /*
-   * TODO -- on some gfx hardware updating the whole texture instead of
-   * the individual rectangles is actually quicker, so we might want to
-   * make this a configurable option (on desktop HW with multiple pipelines
-   * it is usually quicker to just update the damaged parts).
-   *
-   * If we are using TFP we update the whole texture (this simply trigers
-   * the texture rebind).
-   */
-  if (full
-#ifdef HAVE_GLX_TEXTURE_PIXMAP
-      || (CLUTTER_GLX_IS_TEXTURE_PIXMAP (priv->actor) &&
-          clutter_glx_texture_pixmap_using_extension
-                  (CLUTTER_GLX_TEXTURE_PIXMAP (priv->actor)))
-#endif /* HAVE_GLX_TEXTURE_PIXMAP */
-      )
-    {
-      XDamageSubtract (xdisplay, priv->damage, None, None);
-
-      clutter_x11_texture_pixmap_update_area
-	(CLUTTER_X11_TEXTURE_PIXMAP (priv->actor),
-	 0,
-	 0,
-	 clutter_actor_get_width (priv->actor),
-	 clutter_actor_get_height (priv->actor));
-    }
-  else
-    {
-      XRectangle   *r_damage;
-      XRectangle    r_bounds;
-      XserverRegion parts;
-      int           i, r_count;
-
-      parts = XFixesCreateRegion (xdisplay, 0, 0);
-      XDamageSubtract (xdisplay, priv->damage, None, parts);
-
-      r_damage = XFixesFetchRegionAndBounds (xdisplay,
-					     parts,
-					     &r_count,
-					     &r_bounds);
-
-      if (r_damage)
-	{
-	  for (i = 0; i < r_count; ++i)
-	    {
-	      clutter_x11_texture_pixmap_update_area
-		(CLUTTER_X11_TEXTURE_PIXMAP (priv->actor),
-		 r_damage[i].x,
-		 r_damage[i].y,
-		 r_damage[i].width,
-		 r_damage[i].height);
-	    }
-	}
-
-      XFree (r_damage);
-      XFixesDestroyRegion (xdisplay, parts);
-    }
-
   meta_error_trap_pop (display, FALSE);
 
-  priv->needs_repair = FALSE;
+  priv->needs_pixmap = FALSE;
+}
+
+static gboolean
+is_frozen (MutterWindow *self)
+{
+  return self->priv->freeze_count ? TRUE : FALSE;
 }
 
 void
 mutter_window_process_damage (MutterWindow       *self,
 			      XDamageNotifyEvent *event)
 {
-  mutter_window_mark_for_repair (self);
+  MutterWindowPrivate *priv = self->priv;
+  ClutterX11TexturePixmap *texture_x11 = CLUTTER_X11_TEXTURE_PIXMAP (priv->actor);
+
+  if (is_frozen (self))
+    {
+      /* The window is frozen due to an effect in progress: we ignore damage
+       * here on the off chance that this will stop the corresponding
+       * texture_from_pixmap from being update.
+       *
+       * needs_damage_all tracks that some unknown damage happened while the
+       * window was frozen so that when the window becomes unfrozen we can
+       * issue a full window update to cover any lost damage.
+       *
+       * It should be noted that this is an unreliable mechanism since it's
+       * quite likely that drivers will aim to provide a zero-copy
+       * implementation of the texture_from_pixmap extension and in those cases
+       * any drawing done to the window is always immediately reflected in the
+       * texture regardless of damage event handling.
+       */
+      priv->needs_damage_all = TRUE;
+      return;
+    }
+
+
+  clutter_x11_texture_pixmap_update_area (texture_x11,
+                                          event->area.x,
+                                          event->area.y,
+                                          event->area.width,
+                                          event->area.height);
 }
 
 void
@@ -1687,17 +1748,15 @@ mutter_window_update_shape (MutterWindow   *self,
 void
 mutter_window_pre_paint (MutterWindow *self)
 {
-  MutterWindowPrivate *priv = self->priv;
-
-  /* The window is frozen due to a pending animation: we'll wait until
-   * the animation finishes to reshape and repair the window */
-  if (priv->destroy_in_progress    ||
-      priv->maximize_in_progress   ||
-      priv->unmaximize_in_progress)
-    return;
+  if (is_frozen (self))
+    {
+      /* The window is frozen due to a pending animation: we'll wait until
+       * the animation finishes to reshape and repair the window */
+      return;
+    }
 
   check_needs_reshape (self);
-  check_needs_repair (self);
+  check_needs_pixmap (self);
 }
 
 void
